"""Descriptors encryption context"""

import hashlib
from dataclasses import dataclass, field

from luna_cipher.cipher_factory import cipherFactory
from luna_cipher.ciphers.base_cipher import BaseCipher
from luna_cipher.keys_bank.vault_keys_bank import VaultKVKeysBank
from vlutils.helpers import base64ToBytes

from ..constants.derscriptor import ENCRYPTION_HASH_SIZE
from ..utils import mixins


@dataclass
class EncryptionCtx(mixins.Initializable):
    """Descriptor encryption settings"""

    # whether encryption enabled or not
    enabled: bool
    # encryption algorithm
    algorithm: str | None
    # encryption params
    params: dict | None
    # encryption key
    key: bytes | dict | None = field(init=False, default=None)
    # encryption hash (key + algorithm)
    encryptionHash: bytes | None = field(init=False, default=None)
    # cipher implementation
    cipher: BaseCipher | None = field(init=False, default=None)

    async def initialize(self):
        """Initialize"""
        if not self.enabled:
            return
        match self.params["source"]:
            case "vaultKV":
                key, _ = await VaultKVKeysBank(
                    url=self.params["key"]["url"], token=self.params["key"]["token"]
                ).getKey()
            case "raw":
                key = self.params["key"]
            case _:
                raise ValueError(f"Unknown encryption source {self.params['source']}")

        self.key = base64ToBytes(key)
        self.encryptionHash = hashlib.sha256(self.key + self.algorithm.encode()).digest()[:ENCRYPTION_HASH_SIZE]
        self.cipher = cipherFactory(algorithm=self.algorithm)(key=self.key)
