"""
Module contains models for http callbacks
"""

from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any, Coroutine, Literal

from aiohttp import BasicAuth, ClientSession, ClientTimeout
from luna3.common.requests import RequestPayload, makeRequest
from pydantic import ConfigDict, PrivateAttr

from ..utils.log import logger
from .bases import JsonBaseModel as _BaseModel
from .types import Int, OptionalNotNullable, Str


class BaseSchema(_BaseModel):
    """Base schema model"""

    model_config = ConfigDict(ignored_types=(cached_property,))


class BasicAuthorization(BaseSchema):
    """Callback basic authorization"""

    # authorization type
    type: Literal["basic"]
    # authorization login
    login: Str(maxLength=128)
    # authorization password
    password: Str(maxLength=128)

    @property
    def headers(self):
        """Get basic authorization headers"""
        return {"Authorization": BasicAuth(login=self.login, password=self.password, encoding="utf-8").encode()}


class Params(BaseSchema):
    """Callback request parameters"""

    # callback request timeout
    timeout: Int(minValue=0) = OptionalNotNullable()
    # callback request content type
    contentType: Literal["application/json", "application/msgpack"] = OptionalNotNullable()
    # callback request headers
    headers: dict[str, str] = OptionalNotNullable()


class HttpCallback(BaseSchema, ABC):
    """Notification policy callbacks"""

    # callback type
    type: Literal["http"]
    # callback url
    url: str
    # callback authorization parameters
    authorization: BasicAuthorization = OptionalNotNullable()
    # callback request parameters
    params: Params = Params()
    # whether callback enabled or not
    enable: Literal[0, 1] = 1
    # callback final headers
    _headers: dict[str, str] = PrivateAttr()

    def model_post_init(self, __context):
        """Get basic authorization headers"""
        self._headers = self.params.headers or {}
        if self.authorization is not None:
            self._headers |= self.authorization.headers  # pylint: disable-msg=no-member

    @classmethod
    async def onStartup(cls):
        """Init Policies"""
        cls.session = ClientSession(timeout=ClientTimeout(total=60))

    @classmethod
    async def onShutdown(cls):
        """Stop Policies"""
        await cls.session.close()

    @abstractmethod
    def getJsonBody(self, data: Any) -> dict | None:
        """Get json body for sending"""

    def buildCallback(self, data: Any) -> Coroutine | None:
        """
        Build callback coroutine
        Args:
            data: data to send

        Returns:
            callback coroutine or None if callback disabled or filtered
        """
        if not self.enable or not (data4Send := self.getJsonBody(data)):
            return None

        if getattr(self.params, "contentType", None) == "application/msgpack":
            body = RequestPayload.buildMsgpack(data4Send)
        else:
            body = RequestPayload.buildJson(data4Send)

        async def sendNotification():
            reply = await makeRequest(
                url=self.url,
                method="POST",
                headers=self._headers,
                totalTimeout=self.params.timeout,
                body=body,
                session=self.session,
                asyncRequest=True,
            )
            if not reply.success:
                logger.warning(f"Notification to {self.url} failed. {reply.text}")

        return sendNotification()
