"""
"Initializable" components.
"""

from typing import Awaitable, Type

from luna3.sesions import HTTPSession as HTTPSessionBase
from luna3.settings import HTTPSettings
from vlutils.jobs.async_runner import AsyncRunner as AsyncRunnerBase

from . import mixins
from .healthcheck import checkService


class HTTPSession(HTTPSessionBase, mixins.Initializable):
    """Inherit luna3 http session, add "initializable" behavior."""

    def __init__(self, name: str, *args, **kwargs):
        """
        Args:
            name: A str, used to distinguish between sessions in healthcheck output
        """
        super().__init__(*args, **kwargs)
        self.name = name

    async def close(self):
        """Close session."""
        if self._aiohttpSession is None:
            return

        await super().close()

    async def probe(self):
        """Check that we can connect to the service with current configuration."""
        client = self.getClient()
        return await client.testConnection(totalTimeout=1, asyncRequest=True)

    def getRuntimeChecks(self, includeLunaServices: bool = False) -> list[tuple[str, Awaitable]]:
        """
        Returns configured system checks, pairs of (name, coroutine).

        Args:
            includeLunaServices: A bool, whether to return checks for luna services.
        """
        if not includeLunaServices:
            return []

        client = self.getClient()
        return [(self.name, checkService(client, includeLunaServices))]


def makeSession(name: str, apiCls: Type, address: "ServiceAddress", timeouts: "ServiceTimeouts") -> HTTPSession:
    """
    Fabric for creating HTTPSession instance.
    """
    settings = HTTPSettings(
        api=address.apiVersion,
        origin=address.origin,
        totalTimeout=timeouts.totalTimeout,
        connectTimeout=timeouts.connectTimeout,
        sockConnectTimeout=timeouts.sockConnectTimeout,
        sockReadTimeout=timeouts.sockReadTimeout,
        asyncRequest=True,
    )
    return HTTPSession(name, apiCls, settings)


class AsyncRunner(AsyncRunnerBase, mixins.Initializable):
    """Add close() behavior."""
