r"""
===============
CORS Middleware
===============
This is port of https://github.com/playpauseandstop/aiohttp-middlewares for sanic
.. versionadded:: 0.2.0

Dealing with CORS headers for aiohttp applications.

**IMPORTANT:** In same time this CORS middleware mimics the
logic of `django-cors-headers <https://pypi.org/project/django-cors-headers>`_,
where all handling done in the middleware without any additional handlers. This
approach allows aiohttp application to respond with CORS headers for OPTIONS or
wildcard handlers, which is not possible with ``aiohttp-cors`` due to
https://github.com/aio-libs/aiohttp-cors/issues/241 issue.

For detailed information about CORS (Cross Origin Resource Sharing) please
visit:

- `Wikipedia <https://en.m.wikipedia.org/wiki/Cross-origin_resource_sharing>`_
- Or `MDN <https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS>`_

Configuration
=============

**IMPORTANT:** By default, CORS middleware do not allow any origins to access
content from your aiohttp appliction. Which means, you need carefully check
possible options and provide custom values for your needs.
"""

import re
from typing import Any, Collection, Dict, Pattern, Tuple, Type, Union

from sanic.response import HTTPResponse
from yarl import URL

from ..web.request import VLRequest

DictStrAny = Dict[str, Any]
DictStrStr = Dict[str, str]

ExceptionType = Type[Exception]

IntCollection = Collection[int]
StrCollection = Collection[str]

Url = Union[str, Pattern[str], URL]
UrlCollection = Collection[Url]
UrlDict = Dict[Url, StrCollection]
Urls = Union[UrlCollection, UrlDict]


def matchPath(item: Url, path: str) -> bool:
    """Check whether current path is equal to given URL str or regexp.

    Attributes:
        item: URL to compare with request path.
        path: Request path string.
    Returns:
        if url match with path otherwise false
    """
    if isinstance(item, URL):
        item = str(item)

    if isinstance(item, str):
        return item == path

    try:
        return bool(item.match(path))
    except (AttributeError, TypeError):
        return False


def matchItems(items: UrlCollection, value: str) -> bool:
    """Go through all items and try to match item with given value."""
    return any(matchPath(item, value) for item in items)


ACCESS_CONTROL = "Access-Control"
ACCESS_CONTROL_ALLOW = f"{ACCESS_CONTROL}-Allow"
ACCESS_CONTROL_ALLOW_CREDENTIALS = f"{ACCESS_CONTROL_ALLOW}-Credentials"
ACCESS_CONTROL_ALLOW_HEADERS = f"{ACCESS_CONTROL_ALLOW}-Headers"
ACCESS_CONTROL_ALLOW_METHODS = f"{ACCESS_CONTROL_ALLOW}-Methods"
ACCESS_CONTROL_ALLOW_ORIGIN = f"{ACCESS_CONTROL_ALLOW}-Origin"
ACCESS_CONTROL_EXPOSE_HEADERS = f"{ACCESS_CONTROL}-Expose-Headers"
ACCESS_CONTROL_MAX_AGE = f"{ACCESS_CONTROL}-Max-Age"
ACCESS_CONTROL_REQUEST_METHOD = f"{ACCESS_CONTROL}-Request-Method"

# DEFAULT_ALLOW_HEADERS = (
#     "accept",
#     "accept-encoding",
#     "authorization",
#     "content-type",
#     "dnt",
#     "origin",
#     "user-agent",
#     "x-csrftoken",
#     "x-requested-with",
# )
# DEFAULT_ALLOW_METHODS = ("DELETE", "GET", "OPTIONS", "PATCH", "POST", "PUT")
DEFAULT_ALLOW_METHODS = ("*",)
DEFAULT_ALLOW_HEADERS = ("*",)
DEFAULT_URLS: Tuple[Pattern[str]] = (re.compile(r".*"),)
MAX_AGE = 24 * 60 * 60


def patchResponse(
    response: HTTPResponse,
    request: VLRequest,
    *,
    allowAll: bool = True,
    origins: UrlCollection = None,
    exposeHeaders: StrCollection = None,
    allowHeaders: StrCollection = DEFAULT_ALLOW_HEADERS,
    allowMethods: StrCollection = DEFAULT_ALLOW_METHODS,
    allowCredentials: bool = False,
    maxAge: int = MAX_AGE,
) -> HTTPResponse:
    """
    Patch response corresponding CORS rules

    Args:
        response: response
        request: request
        allowAll: When enabled, allow any Origin to access content from your app
                  application. **Please be careful with enabling this option as it may
                  result in security issues for your application.**
        origins: Allow content access for given list of origins. Support supplying
                 strings for exact origin match or regex instances.
        exposeHeaders: List of headers to be exposed with every CORS request.
        allowHeaders: List of allowed headers
        allowMethods: List of allowed methods.
        allowCredentials:  When enabled apply allow credentials header in response, which results
                            in sharing cookies on shared resources. **Please be careful with
                            allowing credentials for CORS requests.**
        maxAge: Access control max age in seconds.
    Returns:
        patched response
    """
    # Is this an OPTIONS request
    isOptionsRequest = request.method == "OPTIONS"
    # Now check origin heaer
    origin = request.headers.get("Origin")
    # Empty origin - do nothing
    if not origin:
        return response

    # Set allow credentials header if necessary
    if allowCredentials:
        response.headers[ACCESS_CONTROL_ALLOW_CREDENTIALS] = "true"

    # Check whether current origin satisfies CORS policy
    if not allowAll and not (origins and matchItems(origins, origin)):
        return response

    # Now start supplying CORS headers
    # First one is Access-Control-Allow-Origin
    if allowAll and not allowCredentials:
        corsOrigin = "*"
    else:
        corsOrigin = origin
    response.headers[ACCESS_CONTROL_ALLOW_ORIGIN] = corsOrigin

    # Then Access-Control-Expose-Headers
    if exposeHeaders:
        response.headers[ACCESS_CONTROL_EXPOSE_HEADERS] = ", ".join(exposeHeaders)

    # Now, if this is an options request, respond with extra Allow headers
    if isOptionsRequest:
        response.headers[ACCESS_CONTROL_ALLOW_HEADERS] = ", ".join(allowHeaders)
        response.headers[ACCESS_CONTROL_ALLOW_METHODS] = ", ".join(allowMethods)
        if maxAge is not None:
            response.headers[ACCESS_CONTROL_MAX_AGE] = str(maxAge)
    return response


async def preCORSMiddleware(request: VLRequest):
    """
    Args:
        request: request
    """

    requestMethod = request.method
    requestPath = request.path

    # Is this an OPTIONS request
    isOptionsRequest = requestMethod == "OPTIONS"

    # Is this a preflight request
    isPreflightRequest = isOptionsRequest and ACCESS_CONTROL_REQUEST_METHOD in request.headers

    # Check whether CORS should be enabled for given URL or not. By default
    # CORS enabled for all URLs
    if not matchItems(DEFAULT_URLS, requestPath):
        return
    if isPreflightRequest:
        response = HTTPResponse()
        return patchResponse(response, request)
    return


async def postCORSMiddleware(request: VLRequest, response: HTTPResponse):
    """
    Middleware, that logs error response.

    Args:
        request: request
        response: response
    """
    patchResponse(response, request)
