"""
Module realizes validators for query params.
"""

from datetime import datetime
from typing import Callable, Iterable, List, Optional, Tuple

from ciso8601 import parse_datetime  # pylint: disable-msg=no-name-in-module
from vlutils.helpers import UTC_TIMEZONE, isUUID

from ..enums.accounts import PermissionsTargets
from ..enums.attributes import ClothingColor, LowerGarmentType, PrimitiveClothingColor, SleeveLength
from ..enums.tasks import TaskStatus, TaskType
from ..errors.errors import ErrorInfo
from ..errors.exception import VLException
from ..errors.pydantic_errors import PydanticError
from ..maps.vl_maps import DEEPFAKE_MAP, EMOTION_MAP, ETHNIC_MAP, LIVENESS_MAP, MASK_MAP
from ..utils.regexps import SPECIAL_SYMBOLS_REGEXP, TRACK_ID_REGEXP

LIST_DELIMITER = ","


def uuidGetter(value: str) -> str:
    """
    Get uuid value.

    Args:
        value: some string
    Raises:
        ValueError: if value is not uuid
    Returns:
    value if  it is uuid
    """
    if isUUID(value):
        return value
    raise ValueError


def timeFilterGetter(value: str) -> datetime:
    """
    Getting time filter. Time  will be converted to UTC without timezone information.

    Args:
        value: time string in isoformat "YYYY-MM-DDThh:mm:ss.sTZD" (eg 1997-07-16T19:20:30.45+01:00) or
               "YYYY-MM-DDThh:mm:ss.sZ"
    Returns:
        datetime object without timezone

    >>> timeFilterGetter("2020-01-01T12:00:00+03:00")
    datetime.datetime(2020, 1, 1, 9, 0)

    >>> timeFilterGetter("2020-01-01T12:00:00Z")
    datetime.datetime(2020, 1, 1, 12, 0)

    >>> timeFilterGetter("2020-01-01T12:00:00+00:00")
    datetime.datetime(2020, 1, 1, 12, 0)

    """
    datetimeValue = parse_datetime(value)
    return datetimeValue.astimezone(UTC_TIMEZONE).replace(tzinfo=None)


def listUUIDsGetter(value: str) -> List[str]:
    """
    Validate list of uuid

    Args:
        value: comma separated uuids
    Returns:
        list of uuid
    >>> listUUIDsGetter("af3fb041-1a55-43e6-9ac2-a361aa432951,98d92aef-8ba9-4b58-a0d9-fcc915160e9d")
    ['af3fb041-1a55-43e6-9ac2-a361aa432951', '98d92aef-8ba9-4b58-a0d9-fcc915160e9d']
    """
    uids = value.split(LIST_DELIMITER)
    for uid in uids:
        uuidGetter(uid)
    return uids


def int01Getter(value: str) -> int:
    """
    Validate int.

    Args:
        value: value 0,1
    Raises:
        ValueError if value not in (0, 1)
    Returns:
        int 0 or 1
    """

    try:
        value = int(value)
    except ValueError as exc:
        raise PydanticError.PydanticValueNotANumberError() from exc
    if value not in (0, 1):
        raise PydanticError.PydanticValidationError.format(f"Value should be either 0 or 1, not {value}")()
    return value


def float01Getter(value: str) -> float:
    """
    Validate float.

    Args:
        value: value 0,1
    Raises:
        ValueError if value not in range between 0 and 1
    Returns:e
        float between 0 and 1
    """

    try:
        value = float(value)
    except ValueError as exc:
        raise PydanticError.PydanticValueNotANumberError() from exc
    if not 0 <= value <= 1:
        raise PydanticError.PydanticValidationError.format(f"Value should be between 1 and 0, not {value}")()
    return value


def floatInRangeGetterFactory(
    minValue: float, maxValue: float, inclusiveMin: Optional[bool] = True, inclusiveMax: Optional[bool] = True
) -> Callable[[str], float]:
    """
    Create float in range validation function

    Args:
        minValue: min value
        maxValue: max value
        inclusiveMin: whether the min value is inclusive or not
        inclusiveMax: whether the max value is inclusive ot not

    Returns:
        query validator function
    """

    def floatInRangeGetter(value: str) -> float:
        """
        Validate float.

        Args:
            value: value

        Raises:
            ValueError if value not in range

        Returns:
            float between minValue and maxValue
        """
        value = float(value)
        if (
            minValue < value < maxValue
            or (value == minValue if inclusiveMin else False)
            or (value == maxValue if inclusiveMax else False)
        ):
            return value
        raise PydanticError.PydanticValidationError.format(
            f"Value should be between {minValue} and {maxValue}, not {value}"
        )()

    return floatInRangeGetter


def enumGetter(inputValues: str, allowedEnum: Tuple) -> List[int]:
    """
    Split string by comma and validate, that items belong to list of allowed integer values
    Args:
        inputValues: comma-separated values
        allowedEnum: tuple with allowed integer values

    Returns:
        list of values
    """
    values = [int(value) for value in inputValues.split(LIST_DELIMITER)]
    if all(value in allowedEnum for value in values):
        return values
    raise ValueError


def emotionsGetter(value: str) -> List[int]:
    """
    Split string by comma and validate, that items belongs to values from EMOTION_MAP that matches to list
     ("anger", "disgust", "fear", "happiness", "neutral", "sadness")

    Args:
        value: comma-separated emotions

    Raises:
        ValueError: if any element from array does not belong to values from EMOTION_MAP exception will be raised

    Returns:
        list of emotions
    """
    return enumGetter(value, EMOTION_MAP.values())


def masksGetter(value: str) -> List[int]:
    """
    Split string by comma and validate, that items belongs to values from MASK_MAP that matches to list
     (1 (missing), 2 (medicalMask), 3 (occluded))

    Args:
        value: comma-separated mask states

    Raises:
        ValueError: if any element from array does not belong to values from MASK_MAP exception will be raised

    Returns:
        list of mask states
    """
    return enumGetter(value, MASK_MAP.values())


def ethnicGroupGetter(value: str) -> List[int]:
    """
    Split string by comma and validate, that items belongs to enum values from ETHNIC_MAP that matches to list
    ("indian", "asian", "african_american", "caucasian")

    Args:
        value: comma-separated ethnic groups

    Raises:
        ValueError: if any element from array does not belong to values from ETHNIC_MAP exception will be raised

    Returns:
        list of emotions
    """
    return enumGetter(value, ETHNIC_MAP.values())


def intRangeGetter(value: str, allowedValues: Iterable[int]) -> List[int]:
    """
    Split string by comma and validate, that items belongs to allowed int values.

    Args:
        value: comma-separated mask states
        allowedValues: allowed values

    Raises:
        ValueError: if any element from array does not belong to allowedValues exception will be raised

    Returns:
        list of parsed int
    """
    values = listIntGetter(value)
    for inputValue in values:
        if inputValue not in allowedValues:
            raise ValueError
    return values


def stringRangeGetter(value: str, allowedValues: Iterable[str]) -> List[str]:
    """
    Split string by comma and validate, that items belongs to allowed string values.

    Args:
        value: comma-separated mask states
        allowedValues: allowed values

    Raises:
        ValueError: if any element from array does not belong to allowedValues exception will be raised

    Returns:
        list of parsed int
    """
    values = listStringsGetter(value)
    for inputValue in values:
        if inputValue not in allowedValues:
            raise ValueError
    return values


ALLOWED_LOWER_GARMENT_TYPES = [garmentType.name for garmentType in LowerGarmentType]


def lowerGarmentGetter(value: str) -> List[str]:
    """
    Split string by comma and validate, that items belongs to enum values from LOWER_GARMENT_MAP that matches to list
    ("trousers", "shorts", "skirt", "undefined")

    Args:
        value: comma-separated lower garment types

    Raises:
        ValueError: if any element from array does not belong to values from LOWER_GARMENT_MAP exception will be raised

    Returns:
        list of lower garment types
    """
    return stringRangeGetter(value, ALLOWED_LOWER_GARMENT_TYPES)


ALLOWED_BACKPACK_STATES = [0, 1, 2]


def backpackStatesGetter(value: str) -> list[int]:
    """Body backpack state filter query getter"""
    return intRangeGetter(value, ALLOWED_BACKPACK_STATES)


ALLOWED_HEADWEAR_STATES = [0, 1, 2]


def headwearStatesGetter(value: str) -> list[int]:
    """Body headwear state filter query getter"""
    return intRangeGetter(value, ALLOWED_HEADWEAR_STATES)


ALLOWED_APPARENT_GENDER = [0, 1, 2]


def apparentGenderGetter(value: str) -> list[int]:
    """Body apparent gender filter query getter"""
    return intRangeGetter(value, ALLOWED_APPARENT_GENDER)


ALLOWED_SLEEVE_LENGTH_VALUES = [sleeve.name for sleeve in SleeveLength]


def sleeveLengthGetter(value: str) -> list[str]:
    """Body sleeve length filter query getter"""
    return stringRangeGetter(value, ALLOWED_SLEEVE_LENGTH_VALUES)


ALLOWED_CLOTHING_COLOR_VALUES = [sleeve.name for sleeve in ClothingColor]


def clothingColorGetter(value: str) -> list[str]:
    """Body upper clothing color filter query getter"""
    return stringRangeGetter(value, ALLOWED_CLOTHING_COLOR_VALUES)


ALLOWED_PRIMITIVE_COLOR_VALUES = [color.name for color in PrimitiveClothingColor]


def primitiveGarmentColorGetter(value: str) -> list[int]:
    """Primitive colors filter query getter"""
    return stringRangeGetter(value, ALLOWED_PRIMITIVE_COLOR_VALUES)


def listStringsGetterCheckSpecialSymbols(value: str) -> List[str]:
    """
    Split string by comma. Also check string consists of characters, numbers, underlines, dashes, commas.

    Args:
        value: string

    Raises:
        ValueError if parameter is not correct

    Returns:
        list of string

    >>> listStringsGetter("1,ab")
    ["1", "ab"]
    """
    if bool(SPECIAL_SYMBOLS_REGEXP.match(value)):
        return value.split(LIST_DELIMITER)
    raise PydanticError.PydanticStringPatternMismatchError.format(SPECIAL_SYMBOLS_REGEXP.pattern)()


def trackIdGetter(value: str) -> str:
    """
    Get and validate trackId

    Args:
        value: A str, value for getting trackId from

    Raises:
        ValueError if trackId is invalid

    Returns:
        A str, valid trackId
    """
    if not TRACK_ID_REGEXP.match(value):
        raise PydanticError.PydanticStringPatternMismatchError.format(TRACK_ID_REGEXP.pattern)()

    return value


def trackIdsGetter(value: str) -> List[str]:
    """
    Split string by comma. Also check string consists of characters, numbers, underlines.
    Args:
        value: string

    Returns:
        list of string
    """
    values = [trackIdGetter(x) for x in value.split(LIST_DELIMITER)]
    return values


def listStringsGetter(value: str) -> List[str]:
    """
    Split string by comma.

    Args:
        value: string

    Returns:
        list of string

    >>> listStringsGetter("1,ab")
    ["1", "ab"]
    """
    return validStringGetter(value).split(LIST_DELIMITER)


def taskTypeGetter(incomingValues: str) -> int:
    """
    Validate task type.

    Args:
        incomingValues: value 0,1 (possible task types)
    Raises:
        ValueError if value not in (0, 1)
    Returns:
        int 0 or 1
    """
    return enumGetter(incomingValues, tuple(taskType.value for taskType in TaskType))[0]


def taskStatusGetter(incomingValues: str) -> int:
    """
    Validate task status.

    Args:
        incomingValues: value 0,1,2,3 (possible task status)
    Raises:
        ValueError if value not in (0, 1, 2, 3)
    Returns:
        int 0 or 1
    """
    return enumGetter(incomingValues, tuple(taskStatus.value for taskStatus in TaskStatus))[0]


def listIntGetter(incomingValues: str) -> List[int]:
    """
    Validate list of int

    Args:
        incomingValues: coma separate int
    Returns:
        list of int
    """
    return list(map(int, incomingValues.split(LIST_DELIMITER)))


def validateIntAsBool(value: int) -> bool:
    """
    Validate int parameter as boolean

    Args:
        value: integer

    Returns:
        boolean
    """
    if int(value) == 1:
        return True
    if int(value) == 0:
        return False
    raise PydanticError.PydanticValidationError.format(f"Value should be either 0 or 1, not {value}")()


def validStringGetter(inputValue: str) -> str:
    """
    Check for empty bytes in a string.

    Args:
        inputValue: string

    Raises:
        ValueError if parameter is not correct

    Returns:
        string
    """
    if "\x00" in inputValue:
        raise PydanticError.PydanticValidationError.format("Value should not contain null symbols")()
    return inputValue


def targetCheckerFactory(
    allowedTargets: List[str], error: ErrorInfo, allowEmpty: bool = False
) -> Callable[[str], List[str]]:
    """
    Create a target-checking function with custom allowed targets.

    Args:
        allowedTargets: allowed targets
        error: error to raise if parameter is not correct
        allowEmpty: whether to allow empty target

    Returns:
        query validator function
    """

    def targetChecker(targetStr: str) -> List[str]:
        """
        Check target query parameter.

        Args:
            targetStr: parameter to validate

        Raises:
            VLException if parameter is not correct

        Returns:
            comma-splitted target
        """
        if not targetStr and allowEmpty:
            return []
        target = targetStr.split(LIST_DELIMITER)
        notAllowedTargets = [column for column in target if column not in allowedTargets]
        if not notAllowedTargets:
            return list(set(target))
        raise VLException(error.format(notAllowedTargets[0], allowedTargets), 400, isCriticalError=False)

    return targetChecker


def boolFrom01Getter(value: str) -> bool:
    """
    Validate bool.

    Args:
        value: value 0,1
    Raises:
        ValueError if value not in (0, 1)
    Returns:
        int 0 or 1
    """
    return bool(int01Getter(value))


def uriParamIntGetter(uriParam: str, error: ErrorInfo, exceededValue: int = 2**32 // 2) -> int:
    """
    Check that the value does not exceed the set value.

    Args:
        uriParam: uri parameter (string representation of an integer)
        error: error to raise if parameter is not correct
        exceededValue: maximum value for uri parameter

    Raises:
        VLException if parameter is not correct

    Returns:
        uri parameter as integer
    """
    if not exceededValue >= int(uriParam) >= -exceededValue:
        raise VLException(error, 404, isCriticalError=False)
    return int(uriParam)


def deepfakeGetter(value: str) -> List[int]:
    """
    Split string by comma and validate, that items belongs to enum values from DEEPFAKE_MAP that matches to list
    (0 ("fake"), 1 ("real"))

    Args:
        value: comma-separated deepfake states

    Raises:
        ValueError: if any element from array does not belong to values from DEEPFAKE_MAP exception will be raised

    Returns:
        list of deepfake state
    """
    return enumGetter(value, DEEPFAKE_MAP.values())


def livenessGetter(value: str) -> List[int]:
    """
    Split string by comma and validate, that items belongs to enum values from LIVENESS_MAP that matches to list
    (0 ("spoof"), 1 ("real"), 2 ("unknown"))

    Args:
        value: comma-separated liveness states

    Raises:
        ValueError: if any element from array does not belong to values from LIVENESS_MAP exception will be raised

    Returns:
        list of liveness state
    """
    return enumGetter(value, LIVENESS_MAP.values())


def userDataGetter(inputValue: str) -> str:
    """
    Check for length and exclude symbol

    Args:
        inputValue: str

    Raises:
        ValueError: if parameter is not correct
        ValueError: if length > 128

    Returns:
        value
    """
    value = validStringGetter(inputValue)
    if len(value) > 128:
        raise PydanticError.PydanticValidationError.format("Value length should be less than or equal 128")()
    return value


def positiveIntegterGetter(value: str) -> int:
    """
    Check for positive integer

    Args:
        inputValue: str

    Raises:
        ValueError: if parameter lower 1 or not integer

    Returns:
        value
    """
    value = int(value)
    if value < 1:
        raise PydanticError.PydanticValidationError.format(f"Value should be greater than 1, not {value}")()
    return value


def tokenPermissionsTargetGetter(value: str) -> PermissionsTargets:
    """
    Validate permissions target.

    Args:
        value: value "standard", "custom", "all" (possible permission targets)
    Raises:
        ValueError if value not in ("standard", "custom", "all")
    Returns:
        string "standard", "custom", "all"
    """
    for target in PermissionsTargets:
        if value == target.name:
            return target
    raise ValueError


def orderGetter(value: str) -> str:
    """
    Raise ValueError if value is not in ("asc", "desc").

    Args:
        value: input query param

    Raises:
        ValueError if parameter is not correct

    Returns:
        value
    """
    if value in ("asc", "desc"):
        return value
    raise PydanticError.PydanticValidationError.format("Value should be either 'asc' or 'desc'")()
